;------------------------------- Port.asm -----------------------------------

;Co-processor communication module using UART.

;Connections:

;Co-processor RXD -> Processor TXD
;Co-processor TXD -> Processor RXD
;Co-processor INT -> Processor INT

;----------------------------- Constantes: ----------------------------------

.equ	Baud	= 500000	;baud rate, bps

.equ UBRRV = (((F_CLK * 10) / (16 * Baud)) - 5) / 10
.if  UBRRV > MAXWORD
.error "out of range constant"
.endif

.equ	C_SET_F	= 0x01	        ;SET FREQUENCY command
.equ	C_SET_P	= 0x02	        ;SET PHASE command
.equ	C_SET_D	= 0x03	        ;SET DELTA PHASE command
.equ	C_SET_A	= 0x04	        ;SET AMPLITUDE command

.equ	CHAN_A	= 4	        ;bit 4 - channel A flag
.equ	CHAN_B	= 5	        ;bit 5 - channel B flag
.equ	CHAN_C	= 6	        ;bit 6 - channel C flag

;----------------------------- Variables: -----------------------------------

.DSEG	;data segment

Comm:	.byte	1		;command code
Buff:	.byte	4		;RX buffer
Bpnt:   .byte   1               ;data pointer

;-------------------------------- Code: -------------------------------------

.CSEG	;Code segment

;----------------------------------------------------------------------------

;Init port:

iPort:  sts     Comm,zero       ;clear command code

;UART0 setup (9 bit):

	sts	UCSRA,zero
	ldi	temp,(1<<RXEN) | (1<<TXEN) | (1<<UCSZ2)
	sts	UCSRB,temp
	ldi	temp,(1<<UCSZ1) | (1<<UCSZ0)
	sts	UCSRC,temp
	ldi	temp,high(UBRRV)
	sts	UBRRH,temp
	ldi	temp,low(UBRRV)
	sts	UBRRL,temp
	ret

;----------------------------------------------------------------------------

;RX frame:

ePort:  lds     cnt,UCSRA       ;read status register
        bbrc    cnt,RXC,exi     ;exit if no received byte
        lds     cnt,UCSRB       ;read RXB8 bit
        lds     temp,UDR        ;read data byte
        Port_CINT_CPL           ;toggle CINT signal
        bbrc    cnt,RXB8,dat    ;if RXB8 == 1 then command

;Command byte received:

cmd:    sts     Comm,temp       ;save command code
        sts     Bpnt,zero       ;clear data pointer
        rjmp    exi

;Data byte received:

dat:    lds     cnt,Comm        ;load command code
        tst     cnt             ;check code
        breq    exi             ;skip unknown data bytes

        lds     cnt,Bpnt        ;load data pointer
        ldx     Buff            ;load buffer base
        add     XL,cnt          ;add offset
        adc     XH,zero
        st      X,temp          ;save data byte
        inc     cnt             ;advance data pointer
        sts     Bpnt,cnt        ;save data pointer
        cpi     cnt,4           ;check pointer
        brne    exi             ;frame not complete

;Execute command:

        lds     temp,Comm       ;load command code
        andi    temp,0x0f       ;select command bits
        cpi     temp,C_SET_F    ;SET FREQUENCY command
        brne    cmd1
        rjmp    SetF
cmd1:   cpi     temp,C_SET_P    ;SET PHASE command
        brne    cmd2
        rjmp    SetP
cmd2:   cpi     temp,C_SET_D    ;SET DELTA PHASE command
        brne    cmd3
        rjmp    SetD
cmd3:   cpi     temp,C_SET_A    ;SET AMPLITUDE command
        brne    cmd4
        rjmp    SetA

cmd4:   sts     Comm,zero       ;clear command code
exi:    ret

;----------------------------------------------------------------------------

;SET FREQUENCY:

SetF:   

#ifdef ALIGN ;FREQUENCY * 2
        ldx     Buff            ;load buffer base
        ld      temp,X
	lsl	temp
	st	X+,temp
        ld      temp,X
	rol	temp
	st	X+,temp
        ld      temp,X
	rol	temp
	st	X+,temp
#endif

	lds     temp,Comm       ;load command code
        StopDDS

        bbrc    temp,CHAN_A,f_b
        ldx     Buff            ;load buffer base
        ldy     AFreq           ;load AFreq base
        ld      cnt,X+          ;set frequency for channel A
        st      Y+,cnt
        ld      cnt,X+
        st      Y+,cnt
        ld      cnt,X+
        st      Y+,cnt

f_b:    bbrc    temp,CHAN_B,f_c
        ldx     Buff            ;load buffer base
        ldy     BFreq           ;load BFreq base
        ld      cnt,X+          ;set frequency for channel B
        st      Y+,cnt
        ld      cnt,X+
        st      Y+,cnt
        ld      cnt,X+
        st      Y+,cnt


f_c:    bbrc    temp,CHAN_C,f_n
        ldx     Buff            ;load buffer base
        ldy     CFreq           ;load CFreq base
        ld      cnt,X+          ;set frequency for channel C
        st      Y+,cnt
        ld      cnt,X+
        st      Y+,cnt
        ld      cnt,X+
        st      Y+,cnt

f_n:    StartDDS
        sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------

;SET PHASE:

SetP:   lds     temp,Comm       ;load command code
        StopDDS

        bbrc    temp,CHAN_A,p_b
        ldx     Buff            ;load buffer base
        ld      APhaseL,X+      ;set phase for channel A
        ld      APhaseM,X+
        ld      APhaseN,X+
        ld      APhaseP,X+

p_b:    bbrc    temp,CHAN_B,p_c
        ldx     Buff            ;load buffer base
        ld      BPhaseL,X+      ;set phase for channel B
        ld      BPhaseM,X+
        ld      BPhaseN,X+
        ld      BPhaseP,X+

p_c:    bbrc    temp,CHAN_C,p_n
        ldx     Buff            ;load buffer base
        ld      CPhaseL,X+      ;set phase for channel C
        ld      CPhaseM,X+
        ld      CPhaseN,X+
        ld      CPhaseP,X+

p_n:    StartDDS
        sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------

;SET DELTA PHASE:

SetD:   StopDDS
        lds     temp,Comm       ;load command code

        bbrc    temp,CHAN_A,d_b
        ldx     Buff            ;load buffer base
        ld      cnt,X+          ;set delta phase for channel A
        add     APhaseL,cnt
        ld      cnt,X+
        adc     APhaseM,cnt
        ld      cnt,X+
        adc     APhaseN,cnt
        ld      cnt,X+
        adc     APhaseP,cnt

d_b:    bbrc    temp,CHAN_B,d_c
        ldx     Buff            ;load buffer base
        ld      cnt,X+          ;set delta phase for channel B
        add     BPhaseL,cnt
        ld      cnt,X+
        adc     BPhaseM,cnt
        ld      cnt,X+
        adc     BPhaseN,cnt
        ld      cnt,X+
        adc     BPhaseP,cnt

d_c:    bbrc    temp,CHAN_C,d_n
        ldx     Buff            ;load buffer base
        ld      cnt,X+          ;set delta phase for channel C
        add     CPhaseL,cnt
        ld      cnt,X+
        adc     CPhaseM,cnt
        ld      cnt,X+
        adc     CPhaseN,cnt
        ld      cnt,X+
        adc     CPhaseP,cnt

d_n:    StartDDS
        sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------

;SET AMPLITUDE:

SetA:   lds     temp,Comm       ;load command code

        bbrc    temp,CHAN_A,a_b
        ldx     Buff            ;load buffer base
        ld      YL,X+           ;set amplitude for channel A
        ld      YH,X+
        sts     OCR0SBH,YH
        sts     OCR0SBL,YL

a_b:    bbrc    temp,CHAN_B,a_c
        ldx     Buff            ;load buffer base
        ld      YL,X+           ;set amplitude for channel B
        ld      YH,X+
        sts     OCR2SBH,YH
        sts     OCR2SBL,YL

a_c:    bbrc    temp,CHAN_C,a_n
        ldx     Buff            ;load buffer base
        ld      YL,X+           ;set amplitude for channel C
        ld      YH,X+
        sts     OCR1SAH,YH
        sts     OCR1SAL,YL

a_n:    sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------
